#!/usr/bin/env python3

import trexio
import numpy as np
from . import nucleus as trexio_nucleus
from . import basis as trexio_basis
from . import ao as trexio_ao
from . import mo as trexio_mo

try:
    import qmckl
    def run(trexio_file, n_points):
        """
        Computes numerically the overlap matrix in the AO basis and compares it to
        the matrix stored in the file.
        """

        trexio_filename = trexio_file.filename
        context = qmckl.context_create()
        qmckl.trexio_read(context, trexio_filename)

        mo = trexio_mo.read(trexio_file)
        ao = mo["ao"]
        basis = ao["basis"]
        nucleus = basis["nucleus"]
        assert basis["type"] == "Gaussian"

        rmin = np.array( list([ np.min(nucleus["coord"][:,a]) for a in range(3) ]) )
        rmax = np.array( list([ np.max(nucleus["coord"][:,a]) for a in range(3) ]) )

        shift = np.array([8.,8.,8.])
        linspace = [ None for i in range(3) ]
        step = [ None for i in range(3) ]
        for a in range(3):
          linspace[a], step[a] = np.linspace(rmin[a]-shift[a], rmax[a]+shift[a], num=n_points, retstep=True)

        print("Integration steps:", step)
        dv = step[0]*step[1]*step[2]

        point = []
        for x in linspace[0]:
          #print(".",end='',flush=True)
          for y in linspace[1]:
            for z in linspace[2]:
               point += [ [x, y, z] ]
        point = np.array(point)
        point_num = len(point)
        mo_num = mo["num"]

        qmckl.set_point(context, 'N', point_num, np.reshape(point, (point_num*3)))
        chi = qmckl.get_mo_basis_mo_value(context, point_num*mo_num)
        chi = np.reshape( chi, (point_num,mo_num) )
        S = chi.T @ chi * dv
        print()

        S_ex = np.eye(mo_num)

        for i in range(mo_num):
          for j in range(i,mo_num):
            print("%3d %3d %15f %15f"%(i,j,S[i][j],S_ex[i,j]))
        S_diff = S - S_ex
        print ("Norm of the error: %f"%(np.linalg.norm(S_diff)))

except ImportError:

    def run(trexio_file, n_points):
        """
        Computes numerically the overlap matrix in the AO basis and compares it to
        the matrix stored in the file.
        """

        mo = trexio_mo.read(trexio_file)
        ao = mo["ao"]
        basis = ao["basis"]
        nucleus = basis["nucleus"]
        assert basis["type"] == "Gaussian"

        rmin = np.array( list([ np.min(nucleus["coord"][:,a]) for a in range(3) ]) )
        rmax = np.array( list([ np.max(nucleus["coord"][:,a]) for a in range(3) ]) )

        shift = np.array([8.,8.,8.])
        linspace = [ None for i in range(3) ]
        step = [ None for i in range(3) ]
        for a in range(3):
          linspace[a], step[a] = np.linspace(rmin[a]-shift[a], rmax[a]+shift[a], num=n_points, retstep=True)

        print("Integration steps:", step)
        dv = step[0]*step[1]*step[2]

        mo_num = mo["num"]
        S = np.zeros( [ mo_num, mo_num ] )
        for x in linspace[0]:
          #print(".",end='',flush=True)
          for y in linspace[1]:
            for z in linspace[2]:
               chi = trexio_mo.value(mo, np.array( [x,y,z] ) )
               S += np.outer(chi, chi)*dv
        print()

        S_ex = np.eye(mo_num)
        S_diff = S - S_ex
        print ("%e %e"%(np.linalg.norm(S), np.linalg.norm(S_ex) ))
        print ("Norm of the error: %e"%(np.linalg.norm(S_diff)))
        #print(S_diff)

        for i in range(mo_num):
          for j in range(i,mo_num):
            print("%3d %3d %15f %15f"%(i,j,S[i][j],S_ex[i,j]))






#!/usr/bin/env python3

import trexio
import numpy as np

def read(trexio_file):
    r = {}

    r["num"] =  trexio.read_nucleus_num(trexio_file)
    r["charge"] =  trexio.read_nucleus_charge(trexio_file)
    r["coord"] =  trexio.read_nucleus_coord(trexio_file)
    r["label"] =  trexio.read_nucleus_label(trexio_file)

    return r






#!/usr/bin/env python3

import trexio
import numpy as np
from . import nucleus

def read(trexio_file):
    r = {}

    r["nucleus"]            =  nucleus.read(trexio_file)

    r["type"]               =  trexio.read_basis_type(trexio_file)
    r["shell_num"]          =  trexio.read_basis_shell_num(trexio_file)
    r["prim_num"]           =  trexio.read_basis_prim_num(trexio_file)
    r["nucleus_index"]      =  trexio.read_basis_nucleus_index(trexio_file)
    r["shell_ang_mom"]      =  trexio.read_basis_shell_ang_mom(trexio_file)
    r["shell_factor"]       =  trexio.read_basis_shell_factor(trexio_file)
    r["shell_index"]        =  trexio.read_basis_shell_index(trexio_file)
    r["exponent"]           =  trexio.read_basis_exponent(trexio_file)
    r["coefficient"]        =  trexio.read_basis_coefficient(trexio_file)
    r["prim_factor"]        =  trexio.read_basis_prim_factor(trexio_file)

    return r


def convert_to_old(basis: dict) -> dict:
    """Convert the new basis set format into the old one (<2.0.0)."""

    basis_old = {}
    # The quantities below did not change in v.2.0.0
    basis_old["type"]               =  basis["type"]
    basis_old["prim_num"]           =  basis["prim_num"]  
    basis_old["shell_ang_mom"]      =  basis["shell_ang_mom"]
    basis_old["shell_factor"]       =  basis["shell_factor"]
    basis_old["exponent"]           =  basis["exponent"]
    basis_old["coefficient"]        =  basis["coefficient"]
    basis_old["prim_factor"]        =  basis["prim_factor"] 
    # basis_num has been renamed into basis_shell_num
    basis_old["num"]                =  basis["shell_num"]
    # The per-nucleus and per-shell lists below have to be reconstructed from the 
    # `nucleus_index` and `shell_index` maps, respectively, introduced in v.2.0.0
    # Save the data in the old format (index of the first shell and No of shells per atom)
    l1, l2 = map_to_lists(basis["nucleus_index"], basis["nucleus"]["num"])
    basis_old["nucleus_index"] = l1 
    basis_old["nucleus_shell_num"] = l2
    # Same for primitives per shell
    l3, l4 = map_to_lists(basis["shell_index"], basis["shell_num"])
    basis_old["shell_prim_index"] = l3 
    basis_old["shell_prim_num"] = l4

    return basis_old


def map_to_lists(map: list, dim: int) -> tuple:
    """Convert long map into two short ones (with index and number of elements per instance (e.g. atom), respectively)."""

    from collections import Counter

    index_per_instance = []
    instances_done = []
    for i, inst in enumerate(map):
        if not inst in instances_done:
            index_per_instance.append(i)
            instances_done.append(inst)

    n_per_instance = Counter(map)
    num_per_instance = [ n_per_instance[j] for j in range(dim) ]

    if(len(index_per_instance) != len(num_per_instance)):
        raise Exception(f"Inconsistent dimensions of output arrays: {len(index_per_instance)} != {len(num_per_instance)}")
    #print(map, "\n", index_per_instance, "\n", num_per_instance)

    return (index_per_instance, num_per_instance)


def lists_to_map(indices: list, numbers: list) -> list:
    """Convert two lists (with index and number of elements per instance like atom) into one big mapping list."""

    if(len(indices) != len(numbers)):
        raise Exception(f"Inconsistent dimensions of input arrays: {len(indices)} != {len(numbers)}")

    map = []
    for i, _ in enumerate(indices):
        for _ in range(numbers[i]):
            map.append(i)
    #print(indices, "\n", numbers, "\n", map)

    return map



#!/usr/bin/env python3
"""
Convert a TREXIO file from one back end into another one.
"""


def generate_converter(json_filename: str) -> None:
    """
    Function that generates the converter based on the trex.json configuration file.
    """
    from os.path import join, dirname, abspath
    from json import load

    try:
        with open(json_filename, 'r') as f:
            config = load(f)
    except FileNotFoundError as e:
        raise Exception(f'File {json_filename} is not found.') from e

    fileDir     = dirname(abspath(__file__))
    output_file = join(fileDir, 'converter_generated.py')

    data_sparse      = []
    data_other       = []
    data_determinant = []
    for group,attr in config.items():
        for data,specs in attr.items():
            name = f'{group}_{data}'
            if 'sparse' in specs[0]:
                data_sparse.append(name)
            elif 'determinant' in group:
                data_determinant.append(name)
            else:
                data_other.append(name)

    with open(output_file, 'w') as f_out:
        f_out.write('import trexio \n')
        f_out.write('def data_handler(trexio_file_from, trexio_file_to) -> None : \n')
        f_out.write('  buffer_size = 20000 \n')

        # Process the normal data first
        for attr in data_other:
            if 'package_version' in attr:
                continue
            block = f'\n\
  if trexio.has_{attr}(trexio_file_from):        \n\
    data = trexio.read_{attr}(trexio_file_from)  \n\
    trexio.write_{attr}(trexio_file_to, data)    \n\
            '
            f_out.write(block)

        # Now process the sparse data
        for attr in data_sparse:
            block = f'\n\
  offset_file = 0 ; eof = False                                   \n\
  if trexio.has_{attr}(trexio_file_from):                         \n\
    while(not eof):                                               \n\
      indices, values, read_size, eof = trexio.read_{attr}(       \n\
          trexio_file_from, offset_file, buffer_size              \n\
          )                                                       \n\
      trexio.write_{attr}(                                        \n\
          trexio_file_to, offset_file, read_size, indices, values \n\
          )                                                       \n\
      offset_file += read_size                                    \n\
            '
            f_out.write(block)

        # Finally process the determinant data
        for attr in data_determinant:
            if 'determinant_num' in attr:
                continue
            block = f'\n\
  offset_file = 0 ; eof = False                                   \n\
  if trexio.has_{attr}(trexio_file_from):                         \n\
    while(not eof):                                               \n\
      data, read_size, eof = trexio.read_{attr}(                  \n\
          trexio_file_from, offset_file, buffer_size              \n\
          )                                                       \n\
      trexio.write_{attr}(                                        \n\
          trexio_file_to, offset_file, read_size, data            \n\
          )                                                       \n\
      offset_file += read_size                                    \n\
            '
            f_out.write(block)


def run_converter(filename_from, filename_to, back_end_to, back_end_from=None, overwrite=None):
    """The high-level converter function."""

    import os
    try:
        import trexio
    except ImportError as exc:
        raise ImportError("trexio Python package is not installed.") from exc

    try:
        # For proper python package
        from converters.converter_generated import data_handler
    except:
        try:
            # For argparse-based CLI
            from converter_generated import data_handler
        except:
            raise ImportError('The generated (JSON-based) data_handler.py module cannot be imported.')

    if not os.path.exists(filename_from):
        raise FileNotFoundError(f'Input file {filename_from} not found.')

    if os.path.exists(filename_to):
        if overwrite:
            if '*' in filename_to:
                raise ValueError('Are you sure?')
            else:
                os.system(f'rm -rf -- {filename_to}')
        else:
            raise ValueError(f'Output file {filename_to} already exists. Consider using the `-w` argument.')

    with trexio.File(filename_from, 'r', back_end_from) as trexio_file_from:
        with trexio.File(filename_to, 'w', back_end_to) as trexio_file_to:
            data_handler(trexio_file_from, trexio_file_to)


def run(filename_from, filename_to, back_end_to, back_end_from=None, overwrite=None, json_filename='trex.json') -> None:
    """Interface to the upstream master script."""

    generate_converter(json_filename)
    run_converter(filename_from, filename_to, back_end_to, back_end_from, overwrite)

    print(f'\n\
        Conversion from {filename_from} to {filename_to} is finished.               \n\
        Note: conversion of the CI coefficients is performed only for ground state. \n\
        ')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Parser')
    parser.add_argument(
        '--json_file',
        type=str,
        nargs=1,
        default='trex.json',
        help='TREXIO configuration file (in JSON format).'
        )
    parser.add_argument(
        '--file_from',
        type=str,
        nargs=1,
        required=True,
        help='Input TREXIO file name.'
        )
    parser.add_argument(
        '--file_to',
        type=str,
        nargs=1,
        required=True,
        help='Output TREXIO file name.'
        )
    parser.add_argument(
        '--back_end_to',
        type=int,
        nargs=1,
        required=True,
        help='Output TREXIO back end.'
        )
    parser.add_argument(
        '--back_end_from',
        type=int,
        nargs=1,
        help='Input TREXIO back end.'
        )
    parser.add_argument(
        '--overwrite',
        type=bool,
        nargs=1,
        default=True,
        help='Overwrite flag. Default: True.'
        )

    args = parser.parse_args()

    json_filename = args.json_file[0]
    filename_from = args.file_from[0]
    filename_to   = args.file_to[0]
    back_end_to   = args.back_end_to[0]
    back_end_from = args.back_end_from[0] if isinstance(args.back_end_from, list) else args.back_end_from
    overwrite     = args.overwrite[0] if isinstance(args.overwrite, list) else args.overwrite

    run(filename_from, filename_to, back_end_to, back_end_from, overwrite, json_filename)



#!/usr/bin/env python3

def to_determinant_list(orbital_list: list, int64_num: int) -> list:
    """Convert an input orbital_list into a set of bit fields."""

    det_list = []
    bitfield = 0
    shift    = 0

    # orbital list has to be sorted in increasing order for the bitfields to be set correctly
    orb_list_sorted = sorted(orbital_list)

    for orb in orb_list_sorted:

        if orb-shift > 64:
            # this removes the 0 bit from the beginning of the bitfield
            bitfield = bitfield >> 1
            # append a bitfield to the list
            det_list.append(bitfield)
            bitfield = 0

        modulo = int((orb-1)/64)
        shift  = modulo*64
        bitfield |= (1 << (orb-shift))

    # this removes the 0 bit from the beginning of the bitfield 
    bitfield = bitfield >> 1
    det_list.append(bitfield)
    #print('Popcounts: ', [bin(d).count('1') for d in det_list)
    #print('Bitfields: ', [bin(d) for d in det_list])
    
    bitfield_num = len(det_list)
    if bitfield_num > int64_num:
        raise Exception(f'Number of bitfields {bitfield_num} cannot be more than the int64_num {int64_num}.') 
    if bitfield_num < int64_num:
        for _ in range(int64_num - bitfield_num):
            print("Appending an empty bitfield.")
            det_list.append(0)

    return det_list




#!/usr/bin/env python3
"""
Set of tools to interact with trexio files.

Usage:
      trexio check-basis      [-n N_POINTS]  [-b BACK_END]  TREXIO_FILE
      trexio check-mos        [-n N_POINTS]  [-b BACK_END]  TREXIO_FILE
      trexio convert-to       -t TYPE -o OUTPUT_FILE        TREXIO_FILE
      trexio convert-from     -t TYPE -i INPUT_FILE  [-b BACK_END]  [-x MO_TYPE]  TREXIO_FILE
      trexio convert-backend  -i INPUT_FILE  -o OUTPUT_FILE  -b BACK_END  -j TREX_JSON_FILE  [-s BACK_END_FROM]  [-w OVERWRITE]
      trexio (-h | --help)

Options:
      -h, --help                    Print help message.
      -n, --n_points=N_POINTS       Number of integration points.  [default: 81]
      -i, --input=INPUT_FILE        Name of the input file.
      -o, --output=OUTPUT_FILE      Name of the output file.
      -b, --back_end=BACK_END       [hdf5 | text | auto]  The TREXIO back end.  [default: hdf5]
      -s, --back_end_from=BACK_END  [hdf5 | text | auto]  The input TREXIO back end.  [default: auto]
      -j, --json=TREX_JSON_FILE     TREX configuration file (in JSON format).
      -w, --overwrite=OVERWRITE     Overwrite flag for the conversion of back ends.  [default: True]
      -t, --type=TYPE               [gaussian | gamess | pyscf | fcidump | molden | cartesian ] File format.
      -x, --motype=MO_TYPE          [natural | initial | guga-initial | guga-natural] The type of the molecular orbitals.
"""

from docopt import docopt
import trexio
import os


def main(filename=None, args=None):
    """Main entry point"""

    if filename is None and args is None:
        args = docopt(__doc__)
        filename = args["TREXIO_FILE"]

    if args["--n_points"]:
       n_points = int(args["--n_points"])
    else:
       n_points = 81

    if args["convert-backend"]:
        if str(args["--overwrite"]).lower() in ['false', '0']:
            overwrite = False
        else:
            overwrite = True
            print(f'File {args["--output"]} will be overwritten.')

    if args["--back_end"]:
        if str(args["--back_end"]).lower() == "hdf5":
            back_end = trexio.TREXIO_HDF5
        elif str(args["--back_end"]).lower() == "text":
            back_end = trexio.TREXIO_TEXT
        elif str(args["--back_end"]).lower() == "auto":
            back_end = trexio.TREXIO_AUTO
        else:
            raise ValueError("Supported back ends: text, hdf5, auto.")
    else:
        if args["convert-backend"]:
            raise Exception("Missing argument for the target back end: specify --back_end or -b.")

    if args["--back_end_from"]:
        if str(args["--back_end_from"]).lower() == "hdf5":
            back_end_from = trexio.TREXIO_HDF5
        elif str(args["--back_end_from"]).lower() == "text":
            back_end_from = trexio.TREXIO_TEXT
        elif str(args["--back_end_from"]).lower() == "auto":
            back_end_from = trexio.TREXIO_AUTO
        else:
            raise ValueError("Supported options: text, hdf5, auto (default).")
    else:
        back_end_from = trexio.TREXIO_AUTO


    if args["check-basis"]:
        trexio_file = trexio.File(filename, 'r', back_end=back_end)
        if trexio_file is None:
            raise IOError

        from .group_tools.check_basis import run
        run(trexio_file,n_points)

    elif args["check-mos"]:
        trexio_file = trexio.File(filename, 'r', back_end=back_end)
        if trexio_file is None:
            raise IOError

        from .group_tools.check_mos import run
        run(trexio_file,n_points)

    elif args["convert-from"]:
        from .converters.convert_from import run
        run(args["TREXIO_FILE"], args["--input"], args["--type"], back_end=back_end, motype=args["--motype"])

    elif args["convert-to"]:
        from .converters.convert_to import run
        run(args["TREXIO_FILE"], args["--output"], args["--type"])

    elif args["convert-backend"]:
        from .converters.convert_back_end import run
        run(
            args["--input"],
            args["--output"],
            back_end_to=back_end,
            back_end_from=back_end_from,
            overwrite=overwrite,
            json_filename=args["--json"]
            )

    else:
        pass


if __name__ == '__main__':
    args = docopt(__doc__)
    filename = args["TREXIO_FILE"]
    main(filename, args)



# pySCF chkpoint file -> TREXIO hdf5 file
# author: Kosuke Nakano
# maintainer: Kosuke Nakano
# email: "kousuke_1123@icloud.com"

# Logger
from logging import getLogger
logger = getLogger("pyscf-trexio").getChild(__name__)


def pyscf_to_trexio(
    pyscf_checkfile: str = "pyscf.chk",
    trexio_filename: str = "trexio.hdf5",
    back_end: str = "hdf5"
):
    """PySCF to TREXIO converter."""

    # load python packages
    import os
    import numpy as np
    # load pyscf packages
    from pyscf import scf
    from pyscf.pbc import scf as pbcscf

    # ## pySCF -> TREX-IO
    # - how to install trexio
    # - pip install trexio

    # import trexio
    import trexio

    logger.info(f"pyscf_checkfile = {pyscf_checkfile}")
    logger.info(f"trexio_filename = {trexio_filename}")
    logger.info("Conversion starts...")

    # pyscf instances
    mol = scf.chkfile.load_mol(pyscf_checkfile)
    mf = scf.chkfile.load(pyscf_checkfile, "scf")

    # PBC info
    try:
        mol.a
        pbc_flag = True
    except AttributeError:
        pbc_flag = False
    logger.info(f"PBC flag = {pbc_flag}")

    # twist_average info
    if pbc_flag:
        try:
            k = mf["kpt"]
            twist_average = False
            logger.info("Single-k calculation")
            k_list = [k]
            if all([k_i == 0.0 for k_i in list(k)]):
                logger.info("k = gamma point")
                logger.info("The generated WF will be real.")
                force_wf_complex = False
            else:
                logger.info("k = general point")
                logger.info("The generated WF will be complex.")
                force_wf_complex = True
        except KeyError:
            twist_average = True
            logger.info("Twisted-average calculation")
            logger.info("Separate TREXIO files are generated")
            logger.info(
                "The correspondence between the index \
                    and k is written in kp_info.dat"
            )
            logger.info("The generated WFs will be complex.")
            force_wf_complex = True
            with open("kp_info.dat", "w") as f:
                f.write("# k_index, kx, ky, kz\n")
            k_list = mf["kpts"]
        finally:
            mol = pbcscf.chkfile.load_cell(pyscf_checkfile)
            k_list = mol.get_scaled_kpts(k_list)
            logger.info(k_list)

    else:
        twist_average = False
        k_list = [[0.0, 0.0, 0.0]]
        force_wf_complex = False

    # if pbc_flag == true, check if ecp or pseudo
    if pbc_flag:
        if len(mol._pseudo) > 0:
            logger.error(
                "TREXIO does not support 'pseudo' format for PBC. Use 'ecp'."
            )
            raise NotImplementedError

    if twist_average:
        logger.warning(
            f"WF at each k point is saved in a separate file kXXXX_{trexio_filename}"
        )
        logger.warning("k points information is stored in kp_info.dat file.")

    # each k WF is stored as a separate file!!
    # for an open-boundary calculation, and a single-k one,
    # k_index is a dummy variable
    for k_index, k_vec in enumerate(k_list):
        assert len(k_vec) == 3  # 3d variable
        # set a filename
        if twist_average:
            logger.info(f"kpt={k_vec}")
            filename = os.path.join(
                os.path.dirname(trexio_filename),
                f"k{k_index}_" + os.path.basename(trexio_filename),
            )
            logger.info(f"filename={filename}")
            with open("kp_info.dat", "a") as f:
                f.write(f"{k_index} {k_vec[0]} {k_vec[1]} {k_vec[2]}\n")
        else:
            filename = trexio_filename

        if os.path.exists(filename):
            logger.warning(f"TREXIO file {filename} already exists and will be removed before conversion.")
            if back_end.lower() == "hdf5":
                os.remove(filename)
            else:
                raise NotImplementedError(f"Please remove the {filename} directory manually.")

        # trexio back end handling
        if back_end.lower() == "hdf5":
            trexio_back_end = trexio.TREXIO_HDF5
        elif back_end.lower() == "text":
            trexio_back_end = trexio.TREXIO_TEXT
        else:
            raise NotImplementedError(f"{back_end} back-end is not supported.")

        # trexio file
        trexio_file = trexio.File(filename, mode="w", back_end=trexio_back_end)

        ##########################################
        # PBC info
        ##########################################
        if pbc_flag:
            Bohr = 0.5291772109
            a = np.array(mol.a[0]) / Bohr  # angstrom -> bohr
            b = np.array(mol.a[1]) / Bohr  # angstrom -> bohr
            c = np.array(mol.a[2]) / Bohr  # angstrom -> bohr
            k_point = k_vec
            periodic = True
        else:
            periodic = False

        # pbc and cell info
        trexio.write_pbc_periodic(trexio_file, periodic)
        if pbc_flag:
            trexio.write_cell_a(trexio_file, a)
            trexio.write_cell_b(trexio_file, b)
            trexio.write_cell_c(trexio_file, c)
            trexio.write_pbc_k_point(trexio_file, k_point)

        # structure info.
        electron_up_num, electron_dn_num = mol.nelec
        nucleus_num = mol.natm
        atom_charges_list = [mol.atom_charge(i) for i in range(mol.natm)]
        """
        atom_nelec_core_list = [
            mol.atom_nelec_core(i) for i in range(mol.natm)
        ]
        atomic_number_list = [
            mol.atom_charge(i) + mol.atom_nelec_core(i)
            for i in range(mol.natm)
        ]
        """
        chemical_symbol_list = [mol.atom_pure_symbol(i) for i in range(mol.natm)]
        atom_symbol_list = [mol.atom_symbol(i) for i in range(mol.natm)]
        coords_np = mol.atom_coords(unit="Bohr")

        ##########################################
        # Structure info
        ##########################################
        trexio.write_electron_up_num(trexio_file, electron_up_num)
        trexio.write_electron_dn_num(trexio_file, electron_dn_num)
        trexio.write_nucleus_num(trexio_file, nucleus_num)
        trexio.write_nucleus_charge(trexio_file, atom_charges_list)
        trexio.write_nucleus_label(trexio_file, chemical_symbol_list)
        trexio.write_nucleus_coord(trexio_file, coords_np)

        ##########################################
        # basis set info
        ##########################################
        # check the orders of the spherical atomic basis in pyscf!!
        # gto.spheric_labels(mol, fmt="%d, %s, %s, %s")
        # for s -> s
        # for p -> px, py, pz
        # for l >= d -> m=(-l ... 0 ... +l)

        basis_type = "Gaussian"  # thanks anthony!
        basis_shell_num = int(np.sum([mol.atom_nshells(i) for i in range(nucleus_num)]))
        nucleus_index = []
        for i in range(nucleus_num):
            for _ in range(len(mol.atom_shell_ids(i))):
                nucleus_index.append(i)
        shell_ang_mom = [mol.bas_angular(i) for i in range(basis_shell_num)]
        basis_prim_num = int(np.sum([mol.bas_nprim(i) for i in range(basis_shell_num)]))

        basis_exponent = []
        basis_coefficient = []
        for i in range(basis_shell_num):
            for bas_exp in mol.bas_exp(i):
                basis_exponent.append(float(bas_exp))
            for bas_ctr_coeff in mol.bas_ctr_coeff(i):
                basis_coefficient.append(float(bas_ctr_coeff))

        basis_shell_index = []
        for i in range(basis_shell_num):
            for _ in range(len(mol.bas_exp(i))):
                basis_shell_index.append(i)

        # normalization factors
        basis_shell_factor = [1.0 for _ in range(basis_shell_num)]  # 1.0 in pySCF

        # gto_norm(l, expnt) => l is angmom, expnt is exponent
        # Note!! Here, the normalization factor of the spherical part
        # are not included. The normalization factor is computed according
        # to Eq.8 of the following paper
        # H.B.S and M.J.F, Int. J. Quant.  Chem., 54(1995), 83-87.
        basis_prim_factor = []
        for prim_i in range(basis_prim_num):
            coeff = basis_coefficient[prim_i]
            expnt = basis_exponent[prim_i]
            l_num = shell_ang_mom[basis_shell_index[prim_i]]
            basis_prim_factor.append(
                mol.gto_norm(l_num, expnt) / np.sqrt(4 * np.pi) * np.sqrt(2 * l_num + 1)
            )

        ##########################################
        # ao info
        ##########################################
        # to be fixed!! for Victor case mol.cart is false, but the basis seems cartesian...
        if mol.cart:
            ao_cartesian = 99999
        else:
            ao_cartesian = 0  # spherical basis representation
        ao_shell = []
        for i, ang_mom in enumerate(shell_ang_mom):
            for _ in range(2 * ang_mom + 1):
                ao_shell.append(i)
        ao_num = len(ao_shell)

        # 1.0 in pyscf (because spherical)
        ao_normalization = [1.0 for _ in range(ao_num)]

        ##########################################
        # mo info
        ##########################################
        mo_type = "MO"

        if twist_average:
            mo_occupation_read = mf["mo_occ"][k_index]
            mo_energy_read = mf["mo_energy"][k_index]
            mo_coeff_read = mf["mo_coeff"][k_index]
        else:
            mo_occupation_read = mf["mo_occ"]
            mo_energy_read = mf["mo_energy"]
            mo_coeff_read = mf["mo_coeff"]

        # check if the pySCF calculation is Restricted or Unrestricted
        # Restricted -> RHF,RKS,ROHF,OROKS
        # Unrestricted -> UHF,UKS

        if len(mo_energy_read) == 2:
            if isinstance(mo_energy_read[0], float):
                spin_restricted = True
            else:
                spin_restricted = False
        else:
            spin_restricted = True

        # the followins are given to TREXIO file lager if spin_restricted == False,
        mo_coefficient_all = []
        mo_occupation_all = []
        mo_energy_all = []
        mo_spin_all = []

        # mo read part starts both for alpha and beta spins
        for ns, spin in enumerate([0, 1]):

            if spin_restricted:
                mo_occupation = mo_occupation_read
                mo_energy = mo_energy_read
                mo_coeff = mo_coeff_read
                if spin == 1:  # 0 is alpha(up), 1 is beta(dn)
                    logger.info("This is spin-restricted calculation.")
                    logger.info("Skip the MO conversion step for beta MOs.")
                    break
            else:
                logger.info(
                    f"MO conversion step for {spin}-spin MOs. 0 is alpha(up), 1 is beta(dn)."
                )
                mo_occupation = mo_occupation_read[ns]
                mo_energy = mo_energy_read[ns]
                mo_coeff = mo_coeff_read[ns]

            mo_num = len(mo_coeff)
            mo_spin_all += [spin for _ in range(mo_num)]

            # mo reordering because mo_coeff[:,mo_i]!!
            mo_coeff = [mo_coeff[:, mo_i] for mo_i in range(mo_num)]

            logger.debug(mo_num)
            logger.debug(len(mo_coeff))
            logger.debug(mo_occupation)
            logger.debug(mo_energy)
            # logger.info(mo_coeff)

            # check if MOs are descending order with respect to "mo occ"
            # this is usually true, but not always true for
            # RO (restricted open-shell) calculations.
            order_bool = all(
                [
                    True if mo_occupation[i] >= mo_occupation[i + 1] else False
                    for i in range(len(mo_occupation) - 1)
                ]
            )
            logger.info(f"MO occupations are in the descending order ? -> {order_bool}")
            if not order_bool:
                logger.warning("MO occupations are not in the descending order!!")
                logger.warning("RO (restricted open-shell) calculations?")
                logger.warning("Reordering MOs...")
                # reordering MOs.
                # descending order (mo occ)
                reo_moocc_index = np.argsort(mo_occupation)[::-1]
                mo_occupation_o = [mo_occupation[l_num] for l_num in reo_moocc_index]
                mo_energy_o = [mo_energy[l_num] for l_num in reo_moocc_index]
                mo_coeff_o = [mo_coeff[l_num] for l_num in reo_moocc_index]
                # descending order (mo energy)
                mo_coeff = []
                mo_occupation = []
                mo_energy = []
                set_mo_occupation = sorted(list(set(mo_occupation_o)), reverse=True)
                for mo_occ in set_mo_occupation:
                    mo_re_index = [
                        i for i, mo in enumerate(mo_occupation_o) if mo == mo_occ
                    ]
                    mo_occupation_t = [mo_occupation_o[l_num] for l_num in mo_re_index]
                    mo_energy_t = [mo_energy_o[l_num] for l_num in mo_re_index]
                    mo_coeff_t = [mo_coeff_o[l_num] for l_num in mo_re_index]
                    reo_ene_index = np.argsort(mo_energy_t)
                    mo_occupation += [mo_occupation_t[l_num] for l_num in reo_ene_index]
                    mo_energy += [mo_energy_t[l_num] for l_num in reo_ene_index]
                    mo_coeff += [mo_coeff_t[l_num] for l_num in reo_ene_index]

            logger.debug("--mo_num--")
            logger.debug(mo_num)
            logger.debug("--len(mo_coeff)--")
            logger.debug(len(mo_coeff))
            logger.debug("--mo_occupation--")
            logger.debug(mo_occupation)
            logger.debug("--mo_energy--")
            logger.debug(mo_energy)
            # logger.debug(mo_coeff)

            # saved mo_occ and mo_energy
            mo_occupation_all += list(mo_occupation)
            mo_energy_all += list(mo_energy)

            # permutation_matrix = []  # for ao and mo swaps, used later

            # molecular coefficient reordering
            # TREX-IO employs (m=-l,..., 0, ..., +l) for spherical basis
            mo_coefficient = []

            for mo_i in range(mo_num):
                mo = mo_coeff[mo_i]
                mo_coeff_buffer = []

                perm_list = []
                perm_n = 0
                for ao_i, ao_c in enumerate(mo):

                    # initialization
                    if ao_i == 0:
                        mo_coeff_for_reord = []
                        current_ang_mom = -1

                    # read ang_mom (i.e., angular momentum of the shell)
                    bas_i = ao_shell[ao_i]
                    ang_mom = shell_ang_mom[bas_i]

                    previous_ang_mom = current_ang_mom
                    current_ang_mom = ang_mom

                    # set multiplicity
                    multiplicity = 2 * ang_mom + 1
                    # print(f"multiplicity = {multiplicity}")

                    # check if the buffer is null, when ang_mom changes
                    if previous_ang_mom != current_ang_mom:
                        assert len(mo_coeff_for_reord) == 0

                    if current_ang_mom == 0:  # s shell
                        # print("s shell/no permutation is needed.")
                        # print("(pyscf notation): s(l=0)")
                        # print("(trexio notation): s(l=0)")
                        reorder_index = [0]

                    elif current_ang_mom == 1:  # p shell

                        # print("p shell/permutation is needed.")
                        # print("(pyscf notation): px(l=+1), py(l=-1), pz(l=0)")
                        # print("(trexio notation): pz(l=0), px(l=+1), py(l=-1)")
                        reorder_index = [2, 0, 1]

                    elif current_ang_mom >= 2:  # > d shell

                        # print("> d shell/permutation is needed.")
                        # print(
                        #    "(pyscf) e.g., f3,-3(l=-3), f3,-2(l=-2), f3,-1(l=-1), \
                        #        f3,0(l=0), f3,+1(l=+1), f3,+2(l=+2), f3,+3(l=+3)"
                        # )
                        # print(
                        #    "(trexio) e.g, f3,0(l=0), f3,+1(l=+1), f3,-1(l=-1), \
                        #        f3,+2(l=+2), f3,-2(l=-2), f3,+3(l=+3), f3,-3(l=-3)"
                        # )
                        l0_index = int((multiplicity - 1) / 2)
                        reorder_index = [l0_index]
                        for i in range(1, int((multiplicity - 1) / 2) + 1):
                            reorder_index.append(l0_index + i)
                            reorder_index.append(l0_index - i)

                    else:
                        raise ValueError("A wrong value was set to current_ang_mom.")

                    mo_coeff_for_reord.append(ao_c)

                    # write MOs!!
                    if len(mo_coeff_for_reord) == multiplicity:
                        # print("--write MOs!!--")
                        mo_coeff_buffer += [
                            mo_coeff_for_reord[i] for i in reorder_index
                        ]

                        # reset buffer
                        mo_coeff_for_reord = []

                        # print("--write perm_list")
                        perm_list += list(np.array(reorder_index) + perm_n)
                        perm_n = perm_n + len(reorder_index)

                mo_coefficient.append(mo_coeff_buffer)
                # permutation_matrix.append(perm_list)

            mo_coefficient_all += mo_coefficient

        # MOs read part end both for alpha and beta spins[l]
        logger.debug("len(mo_coefficient_all)")
        logger.debug(len(mo_coefficient_all))
        logger.debug("len(mo_occupation_all)")
        logger.debug(len(mo_occupation_all))
        logger.debug("len(mo_spin_all)")
        logger.debug(len(mo_spin_all))

        # Conversion from Python complex -> real, complex separately.
        # force WF complex
        if force_wf_complex:
            complex_flag = True
        # check if the MOs have imag.!
        else:
            imag_flags = []
            for mo in mo_coefficient_all:
                imag_flags += list(np.isreal(list(np.real_if_close(mo, tol=100))))
            # print(imag_flags)
            if all(imag_flags):
                complex_flag = False
            else:
                complex_flag = True

        if complex_flag:
            logger.info("The WF is complex")
            mo_coefficient_real = []
            mo_coefficient_imag = []

            for mo__ in mo_coefficient_all:
                mo_real_b = []
                mo_imag_b = []
                for coeff in mo__:
                    mo_real_b.append(coeff.real)
                    mo_imag_b.append(coeff.imag)
                mo_coefficient_real.append(mo_real_b)
                mo_coefficient_imag.append(mo_imag_b)

        else:
            logger.info("The WF is real")
            mo_coefficient_real = [list(np.array(mo).real) for mo in mo_coefficient_all]

        logger.debug("--MOs Done--")

        ##########################################
        # basis set info
        ##########################################
        trexio.write_basis_type(trexio_file, basis_type)  #
        trexio.write_basis_shell_num(trexio_file, basis_shell_num)  #
        trexio.write_basis_prim_num(trexio_file, basis_prim_num)  #
        trexio.write_basis_nucleus_index(trexio_file, nucleus_index)  #
        trexio.write_basis_shell_ang_mom(trexio_file, shell_ang_mom)  #
        trexio.write_basis_shell_factor(trexio_file, basis_shell_factor)  #
        trexio.write_basis_shell_index(trexio_file, basis_shell_index)  #
        trexio.write_basis_exponent(trexio_file, basis_exponent)  #
        trexio.write_basis_coefficient(trexio_file, basis_coefficient)  #
        trexio.write_basis_prim_factor(trexio_file, basis_prim_factor)  #

        ##########################################
        # ao info
        ##########################################
        trexio.write_ao_cartesian(trexio_file, ao_cartesian)  #
        trexio.write_ao_num(trexio_file, ao_num)  #
        trexio.write_ao_shell(trexio_file, ao_shell)  #
        trexio.write_ao_normalization(trexio_file, ao_normalization)  #

        ##########################################
        # mo info
        ##########################################
        trexio.write_mo_type(trexio_file, mo_type)  #

        if complex_flag:
            trexio.write_mo_num(trexio_file, len(mo_coefficient_real))  #
            trexio.write_mo_coefficient(trexio_file, mo_coefficient_real)  #
            trexio.write_mo_coefficient_im(trexio_file, mo_coefficient_imag)  #
        else:
            trexio.write_mo_num(trexio_file, len(mo_coefficient_real))  #
            trexio.write_mo_coefficient(trexio_file, mo_coefficient_real)  #

        trexio.write_mo_occupation(trexio_file, mo_occupation_all)  #

        trexio.write_mo_spin(trexio_file, mo_spin_all)  #

        ##########################################
        # ao integrals
        ##########################################
        # trexio.write_ao_1e_int_overlap(trexio_file, intor_int1e_ovlp)
        # trexio.write_ao_1e_int_kinetic(trexio_file, intor_int1e_kin)
        # trexio.write_ao_1e_int_potential_n_e(trexio_file, intor_int1e_nuc)

        ##########################################
        # ECP
        ##########################################
        # internal format of pyscf
        # https://pyscf.org/pyscf_api_docs/pyscf.gto.html?highlight=ecp#module-pyscf.gto.ecp
        """
        { atom: (nelec,  # core electrons
        ((l, # l=-1 for UL, l>=0 for Ul to indicate |l><l|
        (((exp_1, c_1), # for r^0
        (exp_2, c_2), …),

        ((exp_1, c_1), # for r^1
        (exp_2, c_2), …),

        ((exp_1, c_1), # for r^2
        …))))),

        …}
        """

        # Note! here, the smallest l for the local part is l=1(i.e., p).
        # As a default, nwchem does not have a redundant non-local term
        # (i.e., coeff=0) for H and He.

        if len(mol._ecp) > 0:  # to be fixed!! for Victor case

            ecp_num = 0
            ecp_max_ang_mom_plus_1 = []
            ecp_z_core = []
            ecp_nucleus_index = []
            ecp_ang_mom = []
            ecp_coefficient = []
            ecp_exponent = []
            ecp_power = []

            for nuc_index, (chemical_symbol, atom_symbol) in enumerate(
                zip(chemical_symbol_list, atom_symbol_list)
            ):

                # atom_symbol is superior to atom_pure_symbol!!
                try:
                    z_core, ecp_list = mol._ecp[atom_symbol]
                except KeyError:
                    z_core, ecp_list = mol._ecp[chemical_symbol]

                # ecp zcore
                ecp_z_core.append(z_core)

                # max_ang_mom
                max_ang_mom = max([ecp[0] for ecp in ecp_list])  # this is lmax, right?
                if max_ang_mom == -1:
                    # special case!! H and He.
                    # PySCF database does not define the ul-s part for them.
                    max_ang_mom = 0
                    max_ang_mom_plus_1 = 1
                else:
                    max_ang_mom_plus_1 = max_ang_mom + 1
                ecp_max_ang_mom_plus_1.append(max_ang_mom_plus_1)

                for ecp in ecp_list:
                    ang_mom = ecp[0]
                    if ang_mom == -1:
                        ang_mom = max_ang_mom_plus_1
                    for r, exp_coeff_list in enumerate(ecp[1]):
                        for exp_coeff in exp_coeff_list:
                            exp, coeff = exp_coeff

                            # store variables!!
                            ecp_num += 1
                            ecp_nucleus_index.append(nuc_index)
                            ecp_ang_mom.append(ang_mom)
                            ecp_coefficient.append(coeff)
                            ecp_exponent.append(exp)
                            ecp_power.append(r - 2)

                # special case!! H and He.
                # For the sake of clarity, here I put a dummy coefficient (0.0)
                # for the ul-s part here.
                ecp_num += 1
                ecp_nucleus_index.append(nuc_index)
                ecp_ang_mom.append(0)
                ecp_coefficient.append(0.0)
                ecp_exponent.append(1.0)
                ecp_power.append(0)

            # write to the trex file
            trexio.write_ecp_num(trexio_file, ecp_num)
            trexio.write_ecp_max_ang_mom_plus_1(trexio_file, ecp_max_ang_mom_plus_1)
            trexio.write_ecp_z_core(trexio_file, ecp_z_core)
            trexio.write_ecp_nucleus_index(trexio_file, ecp_nucleus_index)
            trexio.write_ecp_ang_mom(trexio_file, ecp_ang_mom)
            trexio.write_ecp_coefficient(trexio_file, ecp_coefficient)
            trexio.write_ecp_exponent(trexio_file, ecp_exponent)
            trexio.write_ecp_power(trexio_file, ecp_power)

        # close the TREX-IO file
        trexio_file.close()

    logger.info("Conversion to TREXIO is done.")


def cli():
    import argparse
    from logging import getLogger, StreamHandler, Formatter

    log_level = "INFO"
    logger = getLogger("pyscf-trexio")
    logger.setLevel(log_level)
    stream_handler = StreamHandler()
    stream_handler.setLevel(log_level)
    handler_format = Formatter("%(message)s")
    stream_handler.setFormatter(handler_format)
    logger.addHandler(stream_handler)

    # define the parser
    parser = argparse.ArgumentParser(
        epilog="From pyscf chk file to TREXIO file",
        usage="python pyscf_to_trexio.py -c \
            pyscf_checkfile -o trexio_filename",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--pyscf_checkfile",
        help="pyscf checkfile",
        type=str,
        required=True,
    )
    parser.add_argument(
        "-o",
        "--trexio_filename",
        help="trexio filename",
        type=str,
        default="trexio.hdf5",
    )
    parser.add_argument(
        "-b",
        "--back_end",
        help="trexio I/O back-end",
        type=str,
        default="hdf5",
    )

    # parse the input values
    args = parser.parse_args()
    # parsed_parameter_dict = vars(args)

    pyscf_to_trexio(
        pyscf_checkfile=args.pyscf_checkfile,
        trexio_filename=args.trexio_filename,
        back_end=args.back_end
    )


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3

import trexio
import numpy as np
from . import ao as trexio_ao

def read(trexio_file):
    r = {}

    r["ao"]     = trexio_ao.read(trexio_file)
    r["num"]    = trexio.read_mo_num(trexio_file)
    r["coefficient"]  = trexio.read_mo_coefficient(trexio_file)

#    if trexio.has_mo_type(trexio_file):
#        r["type"]   = trexio.read_mo_type(trexio_file)
#    if trexio.has_mo_class(trexio_file):
#        r["class"]   = trexio.read_mo_class(trexio_file)
#    if trexio.has_mo_symmetry(trexio_file):
#        r["symmetry"]   = trexio.read_mo_symmetry(trexio_file)
#    if trexio.has_mo_occupation(trexio_file):
#        r["occupation"]   = trexio.read_mo_occupation(trexio_file)

    return r


def value(mo, r):
    """
    Evaluates all the MOs at R=(x,y,z)
    """

    ao = mo["ao"]
    ao_value = trexio_ao.value(ao, r)
    coef = mo["coefficient"]
    return coef @ ao_value



#!/usr/bin/env python3

import trexio
import numpy as np
from . import basis as trexio_basis

def read(trexio_file):
    r = {}

    r["basis"]     = trexio_basis.read(trexio_file)
    r["basis_old"] = trexio_basis.convert_to_old(r["basis"])
    r["num"]       = trexio.read_ao_num(trexio_file)
    r["shell"]     = trexio.read_ao_shell(trexio_file)
    r["factor"]    = trexio.read_ao_normalization(trexio_file)

    return r

powers = [
 [(0,0,0)],
 [(1,0,0), (0,1,0), (0,0,1)],
 [(2,0,0), (1,1,0), (1,0,1), (0,2,0), (0,1,1), (0,0,2)],
 [(3,0,0), (2,1,0), (2,0,1), (1,2,0), (1,1,1), (1,0,2),
  (0,3,0), (0,2,1), (0,1,2), (0,0,3)]
  ]


def value(ao,r):
    """
    Evaluates all the basis functions at R=(x,y,z)
    """

    basis          =  ao["basis"]

    nucleus        =  basis["nucleus"]
    coord          =  nucleus["coord"]
    nucleus_num    =  nucleus["num"]

    basis_num      =  basis["shell_num"]
    prim_num       =  basis["prim_num"]
    shell_ang_mom  =  basis["shell_ang_mom"]

    # to reconstruct for compatibility with TREXIO < v.2.0.0
    basis_old         = ao["basis_old"]
    nucleus_index     = basis_old["nucleus_index"]
    nucleus_shell_num = basis_old["nucleus_shell_num"]
    shell_prim_index  = basis_old["shell_prim_index"]
    shell_prim_num    = basis_old["shell_prim_num"]

    coefficient = basis["coefficient"] * basis["prim_factor"]
    exponent    = basis["exponent"]

    norm        = ao["factor"]

    # Compute all primitives and powers
    prims = np.zeros(prim_num)
    pows  = [ None for i in range(basis_num) ]

    for i_nucl in range(nucleus_num):

       i_shell = nucleus_index[i_nucl]
       i_prim = shell_prim_index[i_shell]
       istart = i_prim

       try:
         i_shell_end = nucleus_index[i_nucl+1]
         i_prim = shell_prim_index[i_shell_end]
         iend = i_prim
       except IndexError:
         iend = prim_num+1
         i_shell_end = basis_num

       dr = r - coord[i_nucl]
       r2 = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2]
       expo_r = exponent[istart:iend] * r2
       prims[istart:iend] = coefficient[istart:iend] * np.exp(-expo_r)

       old = None
       for i in range(i_shell,i_shell_end):
         if shell_ang_mom[i] != old:
            old = shell_ang_mom[i]
            x = np.array([ np.power(dr, p) for p in powers[old] ])
            x = np.prod(x,axis=1)
         pows[i] = x

    # Compute contractions
    rr = np.zeros(basis_num)
    for i_nucl in range(nucleus_num):
       for i in range(nucleus_shell_num[i_nucl]):
          i_shell = nucleus_index[i_nucl] + i
          n_prim = shell_prim_num[i_shell]
          i_prim = shell_prim_index[i_shell]
          rr[i_shell] = sum(prims[i_prim:i_prim+n_prim])

    result = np.concatenate( [ rr[i] * p for i,p in enumerate(pows) ] )

    return result * norm




#!/usr/bin/env python3
"""
Convert output file from a given code/format into TREXIO
"""

import os
from trexio_tools.group_tools import basis as trexio_basis
from trexio_tools.group_tools import determinant as trexio_det

from .pyscf_to_trexio import pyscf_to_trexio as run_pyscf

try:
    import trexio
except ImportError as exc:
    raise ImportError("trexio Python package is not installed.") from exc

try:
    from resultsFile import getFile, a0, get_lm
    import resultsFile
except ImportError as exc:
    raise ImportError("resultsFile Python package is not installed.") from exc

# Re-order AOs (xx,xy,xz,yy,yz,zz) or (d+0,+1,-1,-2,+2,-2,...)
def f_sort(x):
  if '+' in x or '-' in x:
      l, m = get_lm(x)
      if m>=0:
          return 2*m
      else:
          return -2*m+1
  else:
      return x



def run_resultsFile(trexio_file, filename, motype=None):

    try:
        res = getFile(filename)
    except:
        raise
    else:
        print(filename, 'recognized as', str(res).split('.')[-1].split()[0])

    res.clean_uncontractions()

    # Metadata
    # --------

    trexio.write_metadata_code_num(trexio_file, 1)
    trexio.write_metadata_code(trexio_file,
               [str(res).split('.')[-1].split()[0].replace("File","")] )
    trexio.write_metadata_author_num(trexio_file, 1)
    trexio.write_metadata_author(trexio_file, [res.author])
    trexio.write_metadata_description(trexio_file, res.title)

    # Electrons
    # ---------

    trexio.write_electron_up_num(trexio_file,res.num_alpha)
    trexio.write_electron_dn_num(trexio_file,res.num_beta)

    # Nuclei
    # ------

    charge = []
    coord = []
    nucleus_num = len(res.geometry)

    for a in res.geometry:
        charge.append(a.charge)
        if res.units != 'BOHR':
            coord.append([a.coord[0] / a0, a.coord[1] / a0, a.coord[2] / a0])
        else:
            coord.append([a.coord[0], a.coord[1], a.coord[2]])

    trexio.write_nucleus_num(trexio_file, nucleus_num)
    trexio.write_nucleus_coord(trexio_file, coord)
    # nucleus_charge will be written later after removing core electrons with ECP

    # Transform H1 into H
    import re
    p = re.compile(r'(\d*)$')
    label = [p.sub("", x.name).capitalize() for x in res.geometry]
    trexio.write_nucleus_label(trexio_file, label)

    trexio.write_nucleus_point_group(trexio_file, res.point_group)


    # Basis

    trexio.write_basis_type(trexio_file, "Gaussian")

    # Check whether the basis is Spherical or Cartesian

    cartesian = True
    for b in res.basis:
        if "d+0" in b.sym:
            cartesian = False
            break
        elif "xx" in b.sym:
            break

    # Build basis set
    nucleus_index = []
    nucl_shell_num = []
    shell_ang_mom = []
    shell_prim_num = []
    shell_prim_index = []
    shell_factor = []
    exponent = []
    coefficient = []
    prim_factor = []
    curr_shell = -1
    curr_shell_idx = 0
    ao_shell = []
    prev_idx = None
    geom = [ a.coord for a in res.geometry ]
    try:
      normf = res.normf
    except AttributeError:
      normf=0
    for b in res.basis:
        # Warning: assumes +0 is always 1st of spherical functions
        if ("y" in b.sym) or ("z" in b.sym):
            pass
        elif (not cartesian) and (b.sym not in ["s", "x"]) and ("0" not in b.sym):
            pass
        else:
            curr_shell += 1
            # count the max_ang_mom of a given shell
            if cartesian:
                shell_ang_mom.append(b.sym.count("x"))
            elif b.sym == "s":
                shell_ang_mom.append(0)
            elif b.sym == "x":
                shell_ang_mom.append(1)
            elif "0" in b.sym:
                l, _ = get_lm(b.sym)
                shell_ang_mom.append(l)
            curr_shell_idx = len(exponent)
            shell_prim_index.append(curr_shell_idx)
            shell_prim_num.append(len(b.prim))
            exponent += [x.expo for x in b.prim]
            coefficient += b.coef
            prim_factor += [1./x.norm for x in b.prim]
            if normf == 0:
                shell_factor.append(1./b.norm)
            else:
                shell_factor.append(1.)
            idx = geom.index(b.center)
            if idx != prev_idx:
                nucleus_index.append(curr_shell)
                if len(nucleus_index) > 1:
                    nucl_shell_num.append(nucleus_index[-1]-nucleus_index[-2])

            prev_idx = idx

        ao_shell.append(curr_shell)

    shell_num = curr_shell+1
    prim_num = len(exponent)

    nucl_shell_num.append(shell_num-nucleus_index[-1])

    assert(sum(nucl_shell_num) == shell_num)

    # Fix x,y,z in Spherical (don't move this before basis set detection!)
    if cartesian:
        pass
    else:
        for b in res.basis:
            if b.sym == 'z':
               b.sym = 'p+0'
            elif b.sym == 'x':
               b.sym = 'p+1'
            elif b.sym == 'y':
               b.sym = 'p-1'

    # ========================================================================== #
    # Conversion below is needed to convert arrays according to TREXIO v.2.0
    nucleus_index_per_shell = trexio_basis.lists_to_map(nucleus_index, nucl_shell_num)
    shell_index_per_prim = trexio_basis.lists_to_map(shell_prim_index, shell_prim_num)
    # ========================================================================= #

    # write total number of shell and primitives
    trexio.write_basis_shell_num(trexio_file,shell_num)
    trexio.write_basis_prim_num(trexio_file,prim_num)

    # write mappings to reconstruct per-atom and per-shell quantities
    trexio.write_basis_nucleus_index(trexio_file,nucleus_index_per_shell)
    trexio.write_basis_shell_ang_mom(trexio_file,shell_ang_mom)
    trexio.write_basis_shell_index(trexio_file,shell_index_per_prim)

    # write normalization factor for each shell
    trexio.write_basis_shell_factor(trexio_file,shell_factor)

    # write parameters of the primitives
    trexio.write_basis_exponent(trexio_file,exponent)
    trexio.write_basis_coefficient(trexio_file,coefficient)
    trexio.write_basis_prim_factor(trexio_file,prim_factor)


    # AO
    # --

    #res.convert_to_cartesian()
    trexio.write_ao_cartesian(trexio_file, cartesian)
    trexio.write_ao_num(trexio_file, len(res.basis))
    trexio.write_ao_shell(trexio_file, ao_shell)

    ao_ordering = []
    accu = []
    prev_shell = None

    # Re-order AOs (xx,xy,xz,yy,yz,zz) or (d+0,+1,-1,-2,+2,-2,...)
    for i,b in enumerate(res.basis):
        shell = ao_shell[i]
        if shell != prev_shell:
            accu.sort()
            ao_ordering += accu
            accu = []
        accu += [(f_sort(b.sym), i, b.sym )]
        prev_shell = shell
    accu.sort()
    ao_ordering += accu
    ao_ordering = [ i for (_,i,_) in ao_ordering ]

    # Normalization
    normalization = []
    for i,k in enumerate(ao_ordering):
        b = res.basis[k]
        orig = res.basis[ao_shell.index(ao_shell[k])]
        prim = b.prim
        prim_norm = [ j.norm for j in prim ]
        oprim = orig.prim
        oprim_norm = [ j.norm for j in oprim ]
        accum = 0.
        for i, ci in enumerate(b.coef):
            ci /= prim_norm[i]
            for j, cj in enumerate(orig.coef):
                cj /= oprim_norm[j]
                accum += ci*cj * oprim[i].overlap(oprim[j])
        accum /= orig.norm**2
        normalization.append(accum)
    trexio.write_ao_normalization(trexio_file, normalization)


    # MOs
    # ---

    if motype is None:
        MO_type = res.determinants_mo_type
    else:
        MO_type = motype
    print ("available motypes", res.mo_types)

    allMOs = res.mo_sets[MO_type]
    trexio.write_mo_type(trexio_file, MO_type)

    full_mo_set  = [(allMOs[i].eigenvalue, i) for i in range(len(allMOs))]
    MOindices = [x[1] for x in full_mo_set]

    ## The following commented portion for the future use.
    # try:
    #     closed  = [(allMOs[i].eigenvalue, i) for i in res.closed_mos]
    #     virtual = [(allMOs[i].eigenvalue, i) for i in res.virtual_mos]
    #     active  = [(allMOs[i].eigenvalue, i) for i in res.active_mos]
    # except:
    #     closed  = []
    #     virtual = []
    #     active  = [(allMOs[i].eigenvalue, i) for i in range(len(allMOs))]

    # closed  = [x[1] for x in closed]
    # active  = [x[1] for x in active]
    # virtual = [x[1] for x in virtual]
    # MOindices = closed + active + virtual

    MOs = []
    for i in MOindices:
        MOs.append(allMOs[i])

    mo_num = len(MOindices)
    while len(MOindices) < mo_num:
        MOindices.append(len(MOindices))

    MOmap = list(MOindices)
    for i in range(len(MOindices)):
        MOmap[i] = MOindices.index(i)

    energies = []
    for i in range(mo_num):
        energies.append(MOs[i].eigenvalue)

    MoMatrix = []
    sym0 = [i.sym for i in res.mo_sets[MO_type]]
    sym = [i.sym for i in res.mo_sets[MO_type]]
    for i in range(len(sym)):
        if sym0[i] is None:
            sym[MOmap[i]] = 'A'
        else:
            sym[MOmap[i]] = sym0[i]

    MoMatrix = []
    for i in range(len(MOs)):
        m = MOs[i]
        for j in ao_ordering:
            MoMatrix.append(m.vector[j])


    trexio.write_mo_num(trexio_file, mo_num)
    trexio.write_mo_coefficient(trexio_file, MoMatrix)
    trexio.write_mo_symmetry(trexio_file, sym)

#       TODO: occupations are not always provided in the output file ??
#    if res.occ_num is not None:
#        OccNum = []
#        for i in MOindices:
#           OccNum.append(res.occ_num[MO_type][i])
#    # Not sure about the part below as it might overwrite values from the
#    # previous step !
#        while len(OccNum) < mo_num:
#            OccNum.append(0.)
#        trexio.write_mo_occupation(trexio_file, OccNum)

    lmax = 0
    nucl_charge_remove = []

    nucl_num = len(res.geometry)
    lmax_plus_1_per_atom = []

    map_l = []
    map_nucleus = []

    if res.pseudo:
      ecp_num_total = 0
      ecp_coef_total = []
      ecp_exp_total = []
      ecp_power_total = []
      for ecp in res.pseudo:
          lmax_atomic = ecp['lmax']
          atom = ecp['atom']-1

          lmax_plus_1_per_atom.append(lmax_atomic)

          nucl_charge_remove.append(ecp['zcore'])

          for l in range(lmax_atomic+1):
              l_str = str(l)

              n_per_l = len(ecp[l_str])

              map_nucleus.extend([atom for _ in range(n_per_l) if n_per_l != 0])
              map_l.extend([l for _ in range(n_per_l) if n_per_l != 0])

              ecp_num_total += n_per_l

              coef_per_l = [arr[0] for arr in ecp[l_str]]
              # shift powers by 2 because of the format
              power_per_l = [arr[1]-2 for arr in ecp[l_str]]
              exp_per_l = [arr[2] for arr in ecp[l_str]]

              ecp_coef_total.extend(coef_per_l)
              ecp_power_total.extend(power_per_l)
              ecp_exp_total.extend(exp_per_l)


      # lmax+1 is one higher that the max angular momentum of the core orbital
      # to be removed (per atom)
      trexio.write_ecp_max_ang_mom_plus_1(trexio_file, lmax_plus_1_per_atom)
      # write core charges to be removed
      trexio.write_ecp_z_core(trexio_file, nucl_charge_remove)
      # write total num of ECP elements
      trexio.write_ecp_num(trexio_file, ecp_num_total)
      # write 1-to-1 mapping needed to reconstruct ECPs
      trexio.write_ecp_ang_mom(trexio_file, map_l)
      trexio.write_ecp_nucleus_index(trexio_file, map_nucleus)
      # write ECP quantities in the TREXIO file
      trexio.write_ecp_power(trexio_file, ecp_power_total)
      trexio.write_ecp_coefficient(trexio_file, ecp_coef_total)
      trexio.write_ecp_exponent(trexio_file, ecp_exp_total)


      for i in range(nucl_num):
          charge[i] -= nucl_charge_remove[i]

    # end if res.pseudo:
    trexio.write_nucleus_charge(trexio_file, charge)

    # Determinants
    # ---------

    # resultsFile has non-empty det_coefficients sometimes
    if len(res.det_coefficients[0]) > 1:

        int64_num       = int((mo_num-1)/64) + 1
        determinant_num = len(res.det_coefficients[0])

        # sanity check
        if res.num_states > 1:
            assert determinant_num == len(res.det_coefficients[1])

        # construct the determinant_list of integer bitfields from resultsFile determinants reprsentation
        det_list = []
        for i in range(determinant_num):
            det_tmp      = []
            orb_list_up  = [ orb+1 for orb in res.determinants[i].get("alpha") ]
            det_tmp     += trexio_det.to_determinant_list(orb_list_up, int64_num)
            orb_list_dn  = [ orb+1 for orb in res.determinants[i].get("beta") ]
            det_tmp     += trexio_det.to_determinant_list(orb_list_dn, int64_num)

            det_list.append(det_tmp)

        # write the CI determinants
        offset_file = 0
        trexio.write_determinant_list(trexio_file, offset_file, determinant_num, det_list)

        # write the CI coefficients
        offset_file = 0
        for s in range(res.num_states):
            trexio_file.set_state(s)
            trexio.write_determinant_coefficient(trexio_file, offset_file, determinant_num, res.det_coefficients[s])


    # close the file before leaving
    trexio_file.close()

    print("Conversion to TREXIO format has been completed.")

    return

def run_molden(trexio_file, filename, normalized_basis=True, multiplicity=None, ao_norm=0):
    import numpy as np

    with open(filename, 'r') as f:
        lines = f.readlines()

    if not lines[0].startswith("[Molden Format]"):
        print("File not in Molden format")
        raise TypeError

    title = lines[1].strip()
    atoms = []
    gto = []
    unit = None
    inside = None
    cartesian = True
    sym = []
    ene = []
    spin = []
    occup = []
    mo_coef = []
    mo = []
    for line in lines:
       line = line.strip()
       if line == "":
          continue
       if line.lower().startswith("[atoms]"):
          if "au" in line.lower().split()[1]:
            unit = "au"
          else:
            unit = "angs"
          inside = "Atoms"
          continue
       elif line.upper().startswith("[GTO]"):
          inside = "GTO"
          continue
       elif line.upper().startswith("[MO]"):
          inside = "MO"
          continue
       elif line.startswith("[5d]") \
         or line.startswith("[7f]") \
         or line.startswith("[9g]"):
           cartesian = False
           continue
       elif line.startswith("["):
          inside = None
       if inside == "Atoms":
          buffer = line.split()
          atoms.append( (buffer[0], int(buffer[2]), float(buffer[3]),
                       float(buffer[4]), float(buffer[5])) )
          continue
       elif inside == "GTO":
          gto.append(line)
          continue
       elif inside == "MO":
          in_coef = False
          if line.lower().startswith("sym"):
             sym.append ( line.split('=')[1].strip() )
          elif line.lower().startswith("ene"):
             ene.append ( float(line.split('=')[1].strip()) )
          elif line.lower().startswith("occ"):
             occup.append ( float(line.split('=')[1].strip()) )
          elif line.lower().startswith("spin"):
             if line.split('=')[1].strip().lower == "alpha":
                spin.append(0)
             else:
                spin.append(1)
          else:
             in_coef = True
          if in_coef:
             buffer = line.split()
             mo.append( (int(buffer[0])-1, float(buffer[1])) )
          if not in_coef and len(mo) > 0:
             mo_coef.append(mo)
             mo = []
          continue

    if len(mo) > 0:
       mo_coef.append(mo)

    # Metadata
    # --------

    trexio.write_metadata_code_num(trexio_file, 1)
    trexio.write_metadata_code(trexio_file, ["Molden"])
    trexio.write_metadata_author_num(trexio_file, 1)
    trexio.write_metadata_author(trexio_file, [os.environ["USER"]])
    trexio.write_metadata_description(trexio_file, title)

    # Electrons
    # ---------

    elec_num = int(sum(occup)+0.5)
    if multiplicity is None:
        up_num = 0
        dn_num = 0
        for o in occup:
            if o > 1.0:
                up_num += 1
                dn_num += 1
            elif o == 1.0:
                up_num += 1
    else:
        up_num = (multiplicity-1 + elec_num)/2
        dn_num = elec_num - up_num
    assert (elec_num == up_num + dn_num)
    trexio.write_electron_up_num(trexio_file,up_num)
    trexio.write_electron_dn_num(trexio_file,dn_num)

    # Nuclei
    # ------

    charge = []
    coord = []
    nucleus_num = len(atoms)

    coord = []
    label = []
    for a in atoms:
        charge.append(float(a[1]))
        label.append(a[0])
        if unit != 'au':
            coord.append([a[2] / a0, a[3] / a0, a[4] / a0])
        else:
            coord.append([a[2], a[3], a[4]])

    trexio.write_nucleus_num(trexio_file, len(atoms))
    trexio.write_nucleus_coord(trexio_file, coord)
    trexio.write_nucleus_charge(trexio_file, charge)
    trexio.write_nucleus_label(trexio_file, label)


    # Basis
    # -----

    nucleus_index = []
    shell_ang_mom = []
    shell_index = []
    shell_prim_index = []
    shell_factor = []
    exponent = []
    coefficient = []
    prim_factor = []
    contraction = None

    shell_id = -1
    prim_id = -1
    iatom = 0
    for line in gto:
       buffer = line.replace('D','E').split()
       if len(buffer) == 2 and buffer[1] == "0":
           iatom = int(buffer[0])-1
       elif len(buffer) == 3 and float(buffer[2]) == 1.0:
           if contraction is not None:
               if normalized_basis:
                    accum = 0.
                    n = [ x.norm for x in contraction.prim ]
                    for i, ci in enumerate(contraction.coef):
                        ci /= n[i]
                        for j, cj in enumerate(contraction.coef):
                            cj /= n[j]
                            accum += ci*cj * contraction.prim[i].overlap(contraction.prim[j])
                    shell_factor.append(1./accum)
               else:
                    shell_factor.append(1.)
           shell_id += 1
           ang_mom = buffer[0].lower()
           nprim = int(buffer[1])
           nucleus_index.append(iatom)
           if   ang_mom == "s": shell_ang_mom.append(0)
           elif ang_mom == "p": shell_ang_mom.append(1)
           elif ang_mom == "d": shell_ang_mom.append(2)
           elif ang_mom == "f": shell_ang_mom.append(3)
           elif ang_mom == "g": shell_ang_mom.append(4)
           if   ang_mom != "s": ang_mom = "x"*shell_ang_mom[-1]
           contraction = resultsFile.contraction()
       else:
           prim_id += 1
           e, c = float(buffer[0]), float(buffer[1])
           shell_prim_index.append(prim_id)
           exponent.append(e)
           coefficient.append(c)
           gauss = resultsFile.gaussian()
           gauss.center = coord[iatom]
           gauss.expo = e
           gauss.sym  =ang_mom
           contraction.append(c, gauss)
           prim_factor.append(1./gauss.norm)
           shell_index.append(shell_id)

    if contraction is not None:
        if normalized_basis:
            accum = 0.
            n = [ x.norm for x in contraction.prim ]
            for i, ci in enumerate(contraction.coef):
                ci /= n[i]
                for j, cj in enumerate(contraction.coef):
                    cj /= n[j]
                    accum += ci*cj * contraction.prim[i].overlap(contraction.prim[j])
            shell_factor.append(1./accum)
        else:
            shell_factor.append(1.)

    shell_num = shell_id + 1
    prim_num  = prim_id  + 1

    trexio.write_basis_type(trexio_file, "Gaussian")

    # write total number of shell and primitives
    trexio.write_basis_shell_num(trexio_file,shell_num)
    trexio.write_basis_prim_num(trexio_file,prim_num)

    # write mappings to reconstruct per-atom and per-shell quantities
    trexio.write_basis_nucleus_index(trexio_file,nucleus_index)
    trexio.write_basis_shell_ang_mom(trexio_file,shell_ang_mom)
    trexio.write_basis_shell_index(trexio_file,shell_index)

    # write normalization factor for each shell
    trexio.write_basis_shell_factor(trexio_file,shell_factor)

    # write parameters of the primitives
    trexio.write_basis_exponent(trexio_file,exponent)
    trexio.write_basis_coefficient(trexio_file,coefficient)
    trexio.write_basis_prim_factor(trexio_file,prim_factor)


    # AOs
    # ---

    if max(shell_ang_mom) < 2:
       cartesian=True

    if cartesian:
        conv = [ [ 's' ], ['x', 'y', 'z'], ['xx', 'yy', 'zz', 'xy', 'xz', 'yz'],
                 ['xxx', 'yyy', 'zzz', 'xyy', 'xxy', 'xxz', 'xzz', 'yzz', 'yyz', 'xyz'],
                 ['xxxx', 'yyyy', 'zzzz', 'xxxy', 'xxxz', 'xyyy', 'yyyz', 'xzzz', 'yzzz',
                  'xxyy', 'xxzz', 'yyzz', 'xxyz', 'xyyz', 'xyzz'] ]
    else:
        conv = [ ['s'], ['p+1', 'p-1', 'p+0'], ['d+0', 'd+1', 'd-1', 'd+2', 'd-2'],
                 ['f+0', 'f+1', 'f-1', 'f+2', 'f-2', 'f+3', 'f-3'],
                 ['g+0', 'g+1', 'g-1', 'g+2', 'g-2', 'g+3', 'g-3', 'g+4', 'g-4'] ]

    norm = []
    for l in range(5):
       g = resultsFile.gaussian()
       gauss.center = (0.,0.,0.)
       gauss.expo = 1.0
       gauss.sym = conv[l][0]
       ref = gauss.norm
       norm.append([])
       for m in conv[l]:
            g = resultsFile.gaussian()
            gauss.center = (0.,0.,0.)
            gauss.expo = 1.0
            gauss.sym = m
            norm[l].append ( gauss.norm / ref )

    ao = []
    ao_normalization = []
    for l in shell_ang_mom:
       if l>4:
          raise TypeError("Angular momentum too high: l>4 not supported by Molden format.")
       ao.append(conv[l])
       ao_normalization.append(norm[l])

    ao_shell = []
    ao_ordering = []
    j = 0
    for k,l in enumerate(ao):
      ao_shell += [ k for _ in l ]
      accu = [ (f_sort(x), i+j, norm[shell_ang_mom[k]][i])  for i,x in enumerate(l) ]
      accu.sort()
      ao_ordering += accu
      j += len(l)
    ao_normalization = [ i for (_,_,i) in ao_ordering ]
    ao_ordering = [ i for (_,i,_) in ao_ordering ]
    ao_num = len(ao_ordering)

    trexio.write_ao_num(trexio_file, ao_num)
    trexio.write_ao_cartesian(trexio_file, cartesian)
    trexio.write_ao_shell(trexio_file, ao_shell)
    trexio.write_ao_normalization(trexio_file, ao_normalization)

    # MOs
    # ---

#    trexio.write_mo_type(trexio_file, MO_type)

    core   = []
    active = []
    virtual = []
    mo_class = []
    for i, o in enumerate(occup):
       if o >= 2.:
          core.append(i)
          mo_class.append("Core")
       elif o == 0.:
          virtual.append(i)
          mo_class.append("Virtual")
       else:
          active.append(i)
          mo_class.append("Active")

    trexio.write_mo_num(trexio_file, len(mo_class))
    MoMatrix = []
    for mo in mo_coef:
      vector = np.zeros(ao_num)
      for i, x in mo:
         vector[i] = x
      for i in ao_ordering:
         MoMatrix.append(vector[i])

    trexio.write_mo_spin(trexio_file, spin)
    trexio.write_mo_class(trexio_file, mo_class)
    trexio.write_mo_occupation(trexio_file, occup)
    trexio.write_mo_symmetry(trexio_file, sym)
    trexio.write_mo_coefficient(trexio_file, MoMatrix)


def run(trexio_filename, filename, filetype, back_end, motype=None):

    if os.path.exists(trexio_filename):
        print(f"TREXIO file {trexio_filename} already exists and will be removed before conversion.")
        if back_end == trexio.TREXIO_HDF5:
            os.remove(trexio_filename)
        else:
            raise NotImplementedError(f"Please remove the {trexio_filename} directory manually.")

    if "pyscf" not in filetype.lower():
        trexio_file = trexio.File(trexio_filename, mode='w', back_end=back_end)

    if filetype.lower() == "gaussian":
        run_resultsFile(trexio_file, filename, motype)
    elif filetype.lower() == "gamess":
        run_resultsFile(trexio_file, filename, motype)
    elif filetype.lower() == "pyscf":
        back_end_str = "text" if back_end==trexio.TREXIO_TEXT else "hdf5"
        run_pyscf(trexio_filename=trexio_filename, pyscf_checkfile=filename, back_end=back_end_str)
    elif filetype.lower() == "fcidump":
        raise NotImplementedError(f"Conversion from {filetype} to TREXIO is not supported.")
        #run_fcidump(trexio_file, filename)
    elif filetype.lower() == "molden":
        run_molden(trexio_file, filename)
    else:
        raise NotImplementedError(f"Conversion from {filetype} to TREXIO is not supported.")



#!/usr/bin/env python3
"""
convert output of GAMESS/GAU$$IAN to trexio
"""

import sys
import os
from functools import reduce
from . import cart_sphe as cart_sphe
import numpy as np

try:
    import trexio
except:
    print("Error: The TREXIO Python library is not installed")
    sys.exit(1)





def run_molden(t, filename):

    out = ["[Molden Format]"]
    out += ["Converted from TREXIO"]

    out += ["[Atoms] AU"]

    nucl_num = trexio.read_nucleus_num(t)
    charge = trexio.read_nucleus_charge(t)
    if trexio.has_ecp_z_core(t):
       z_core = trexio.read_ecp_z_core(t)
       charge = [ x + y for x,y in zip(charge,z_core) ]
    name   = trexio.read_nucleus_label(t)
    coord  = trexio.read_nucleus_coord(t)
    for i in range(nucl_num):
        out += [ "%3s %4d %4d %18.14f %18.14f %18.14f"%tuple(
                 [name[i], i+1, int(charge[i])] + list(coord[i])  ) ]



    basis_type = trexio.read_basis_type(t)
    if basis_type.lower() == "gaussian":

        out += ["[GTO]"]
        prim_num = trexio.read_basis_prim_num(t)
        shell_num = trexio.read_basis_shell_num(t)
        nucleus_index = trexio.read_basis_nucleus_index(t)
        shell_ang_mom = trexio.read_basis_shell_ang_mom(t)
        shell_factor = trexio.read_basis_shell_factor(t)
        shell_index = trexio.read_basis_shell_index(t)
        exponent = trexio.read_basis_exponent(t)
        coefficient = trexio.read_basis_coefficient(t)
        prim_factor = trexio.read_basis_prim_factor(t)

        contr = [ { "exponent"      : [],
                    "coefficient"   : [],
                    "prim_factor"   : []  }  for _ in range(shell_num) ]
        for j in range(prim_num):
            i = shell_index[j]
            contr[i]["exponent"]    += [ exponent[j] ]
            contr[i]["coefficient"] += [ coefficient[j] ]
            contr[i]["prim_factor"] += [ prim_factor[j] ]

        basis = {}
        for k in range(nucl_num):
            basis[k] = { "shell_ang_mom" : [],
                        "shell_factor"  : [],
                        "shell_index"   : [],
                        "contr"         : [] }

        for i in range(shell_num):
            k = nucleus_index[i]
            basis[k]["shell_ang_mom"] += [ shell_ang_mom[i] ]
            basis[k]["shell_factor"]  += [ shell_factor[i] ]
            basis[k]["shell_index"]   += [ shell_index[i] ]
            basis[k]["contr"]         += [ contr[i] ]


        ang_mom_conv = [ "s", "p", "d", "f", "g" ]
        for k in range(nucl_num):
          out += [ "%6d  0"%(k+1) ]
          for l in range(len(basis[k]["shell_index"])):
              ncontr = len(basis[k]["contr"][l]["exponent"])
              out += [ "%2s %8d 1.00" % (
                        ang_mom_conv[ basis[k]["shell_ang_mom"][l] ],
                        ncontr) ]
              for j in range(ncontr):
                out += [ "%20.10E    %20.10E"%(
                          basis[k]["contr"][l]["exponent"][j],
                          basis[k]["contr"][l]["coefficient"][j] ) ]
          out += [""]

    # end if basis_type.lower() == "gaussian"



#   5D: D 0, D+1, D-1, D+2, D-2
#   6D: xx, yy, zz, xy, xz, yz
#
#   7F: F 0, F+1, F-1, F+2, F-2, F+3, F-3
#  10F: xxx, yyy, zzz, xyy, xxy, xxz, xzz, yzz, yyz, xyz
#
#   9G: G 0, G+1, G-1, G+2, G-2, G+3, G-3, G+4, G-4
#  15G: xxxx yyyy zzzz xxxy xxxz yyyx yyyz zzzx zzzy,
#       xxyy xxzz yyzz xxyz yyxz zzxy

    mo_num = trexio.read_mo_num(t)
    cartesian = trexio.read_ao_cartesian(t)
    if cartesian:
      order = [ [0],
                [0, 1, 2],
                [0, 3, 5, 2, 3, 4],
                [0, 6, 9, 3, 1, 2, 5, 8, 7, 4],
                [0, 10, 14, 1, 2, 6, 11, 9, 13, 3, 5, 12, 4, 7, 8] ]
    else:
       out += [ "[5D]", "[7F]", "[9G]" ]
       order = [ [0], [1, 2, 0],
                 [ i for i in range(5) ],
                 [ i for i in range(7) ],
                 [ i for i in range(9) ] ]

    ao_num = trexio.read_ao_num(t)
    o = []
    icount = 0
    for i in range(shell_num):
       l = shell_ang_mom[i]
       for k in order[l]:
          o.append( icount+k )
       icount += len(order[l])


    elec_alpha_num = trexio.read_electron_up_num(t)
    elec_beta_num  = trexio.read_electron_dn_num(t)
    if trexio.has_mo_occupation(t):
       occ = trexio.read_mo_occupation(t)
    else:
       occ = [ 0. for i in range(mo_num) ]
       for i in range(elec_alpha_num):
         occ[i] += 1.
       for i in range(elec_beta_num):
         occ[i] += 1.

    mo_coef = trexio.read_mo_coefficient(t)
    if trexio.has_mo_symmetry(t):
      sym = trexio.read_mo_symmetry(t)
    else:
      sym = [ "A1" for _ in range(mo_num) ]


    out += ["[MO]"]

    for i in range(mo_num):
       out += [ "Sym= %s"%(sym[i]),
#               "Ene= 0.0",
                "Spin= Alpha",
                "Occup= %f"%(occ[i]) ]
       for k in range(ao_num):
           out += [ "%6d  %20.10E"%(k+1, mo_coef[i,o[k]]) ]

    if trexio.has_ecp_z_core(t):
       out += [ "[CORE]" ]
       for i,k in enumerate(z_core):
           out += [ "%4d : %4d"%(i, k) ]

    out += [ "" ]
    out_file = open(filename,"w")
    out_file.write("\n".join(out))

    return




def run_cart_phe(inp, filename, to_cartesian):
    out = trexio.File(filename, 'u', inp.back_end)

    shell_ang_mom = trexio.read_basis_shell_ang_mom(inp)

    # Build transformation matrix
    count_sphe = 0
    count_cart = 0
    accu = []
    shell = []
    for i, l in enumerate(shell_ang_mom):
      p, r = count_cart, count_sphe
      (x,y) = cart_sphe.data[l].shape
      count_cart += x
      count_sphe += y
      q, s = count_cart, count_sphe
      accu.append( (l, p,q, r,s) )
      if to_cartesian != 0: n = x
      else: n = y
      for _ in range(n):
          shell.append(i)

    cart_normalization = np.ones(count_cart)
    R = np.zeros( (count_cart, count_sphe) )
    for (l, p,q, r,s) in accu:
      R[p:q,r:s] = cart_sphe.data[l]
      cart_normalization[p:q] = cart_sphe.normalization[l]

    S = np.eye(count_sphe)

    ao_num_in  = trexio.read_ao_num(inp)


    if to_cartesian == 0:
        print("Transformation from cartesian to spherical is not implemented")
        return

#        # See http://users.df.uba.ar/dmitnik/estructura3/bases/biblio/transformationGaussianSpherical.pdf
#        # If S is the overlap of cartesian functions, R @ S @ R.T = I, so R^{-1} = S @ R.T
#        if trexio.has_ao_1e_int_overlap(inp):
#            X = trexio.read_ao_1e_int_overlap(inp)
#        else:
#            print("Transformation from Cartesian to Spherical requires AO overlap matrix.")
#            return -1
#
#        S = np.zeros((count_cart, count_cart))
#        for (l, p,q, r,s) in accu:
#            S[p:q,p:q] = X[p:q,p:q]
#
#        R = R.T @ S

    elif to_cartesian == -1:
        R = np.eye(ao_num_in)

    # Update AOs
    ao_num_out = R.shape[0]
    trexio.write_ao_cartesian(out, to_cartesian)
    trexio.write_ao_num(out, ao_num_out)
    trexio.write_ao_shell(out, shell)

    normalization = np.array( [ 1. ] * ao_num_in )
    if trexio.has_ao_normalization(inp):
      normalization = trexio.read_ao_normalization(inp)

    trexio.write_ao_normalization(out, cart_normalization)

    R_norm_inv = np.array(R)

    # Update MOs
    if trexio.has_mo_coefficient(inp):
      X = trexio.read_mo_coefficient(inp)
      for i in range(R.shape[1]):
         if normalization[i] != 1.:
            X[:,i] *= normalization[i]
      Y  = X @ R.T
      trexio.write_mo_coefficient(out, Y)

    # Update 1e Integrals
    if trexio.has_ao_1e_int_overlap(inp):
      X = trexio.read_ao_1e_int_overlap(inp)
      for i in range(R.shape[1]):
            X[:,i] /= normalization[i]
      for i in range(R.shape[1]):
            X[i,:] /= normalization[i]
      Y = R_norm_inv @ X @ R_norm_inv.T
      trexio.write_ao_1e_int_overlap(out, Y)


    if trexio.has_ao_1e_int_kinetic(inp):
      X = trexio.read_ao_1e_int_kinetic(inp)
      for i in range(R.shape[1]):
         if normalization[i] != 1.:
            X[:,i] /= normalization[i]
            X[i,:] /= normalization[i]
      trexio.write_ao_1e_int_kinetic(out, R_norm_inv @ X @ R_norm_inv.T)

    if trexio.has_ao_1e_int_potential_n_e(inp):
      X = trexio.read_ao_1e_int_potential_n_e(inp)
      for i in range(R.shape[1]):
         if normalization[i] != 1.:
            X[:,i] /= normalization[i]
            X[i,:] /= normalization[i]
      trexio.write_ao_1e_int_potential_n_e(out, R_norm_inv @ X @ R_norm_inv.T)

    if trexio.has_ao_1e_int_ecp(inp):
      X = trexio.read_ao_1e_int_ecp(inp)
      for i in range(R.shape[1]):
         if normalization[i] != 1.:
            X[:,i] /= normalization[i]
            X[i,:] /= normalization[i]
      trexio.write_ao_1e_int_ecp(out, R_norm_inv @ X @ R_norm_inv.T)

    if trexio.has_ao_1e_int_core_hamiltonian(inp):
      X = trexio.read_ao_1e_int_core_hamiltonian(inp)
      for i in range(R.shape[1]):
         if normalization[i] != 1.:
            X[:,i] /= normalization[i]
            X[i,:] /= normalization[i]
      trexio.write_ao_1e_int_core_hamiltonian(out, R_norm_inv @ X @ R_norm_inv.T)

    # Remove 2e integrals: too expensive to transform
    if trexio.has_ao_2e_int_eri(inp):
      print("Warning: Two-electron integrals are not converted")
      trexio.delete_ao_2e_int_eri(out)

    if trexio.has_ao_2e_int_eri_lr(inp):
      trexio.delete_ao_2e_int_eri_lr(out)


def run_normalized_aos(t, filename):
    # Start by copying the file
    os.system('cp -r %s %s' % (t.filename, filename))
    run_cart_phe(t, filename, to_cartesian=-1)
    return


def run_cartesian(t, filename):
    # Start by copying the file
    os.system('cp -r %s %s' % (t.filename, filename))
    cartesian = trexio.read_ao_cartesian(t)
    if cartesian > 0:
        return

    run_cart_phe(t, filename, to_cartesian=1)
    return

def run_spherical(t, filename):
    # Start by copying the file
    os.system('cp -r %s %s' % (t.filename, filename))
    cartesian = trexio.read_ao_cartesian(t)
    if cartesian == 0:
        return

    run_cart_phe(t, filename, to_cartesian=0)
    return


def run(trexio_filename, filename, filetype):

    trexio_file = trexio.File(trexio_filename,mode='r',back_end=trexio.TREXIO_AUTO)

    if filetype.lower() == "molden":
        run_molden(trexio_file, filename)
    elif filetype.lower() == "cartesian":
        run_cartesian(trexio_file, filename)
    elif filetype.lower() == "spherical":
        run_spherical(trexio_file, filename)
#    elif filetype.lower() == "normalized_aos":
#        run_normalized_aos(trexio_file, filename)
#    elif filetype.lower() == "fcidump":
#        run_fcidump(trexio_file, filename)
    else:
        raise NotImplementedError(f"Conversion from TREXIO to {filetype} is not supported.")

